##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

from __future__ import absolute_import

import os.path as op
import json
import time
import logging
from .httpconn import HttpConn
from .config import Config


class Folder:

    """
    Represents a folder of domains
    """

    @property
    def domain(self):
        domain = self._domain
        if domain is None:
            domain = ""

        r = domain + "/"

        return r

    @property
    def parent(self):
        if self._domain is None:
            return None
        else:
            return op.dirname(self._domain) + "/"

    @property
    def modified(self):
        """Last modified time of the domain as a datetime object."""
        return self._modified

    @property
    def created(self):
        """creation time of the domain as a datetime object."""
        return self._created

    @property
    def owner(self):
        """Username of the owner of the folder"""
        return self._owner

    @property
    def is_folder(self):
        """is this a proper folder (i.e. domain without root group),
        or a domain"""
        if self._obj_class == "folder":
            return True
        else:
            return False

    def __init__(
        self,
        domain_name,
        pattern=None,
        query=None,
        mode=None,
        endpoint=None,
        verbose=False,
        username=None,
        password=None,
        bucket=None,
        api_key=None,
        logger=None,
        owner=None,
        batch_size=1000,
        retries=3,
        **kwds,
    ):
        """Create a new Folders object.

        domain_name
            URI of the domain name to access. E.g.: /org/hdfgroup/folder/
        pattern
            Regex filter to apply to any domains in the folder
        query
            Query for root level attributes - only domains matching the query will be returned
        mode
            Access mode: 'r', 'r+', 'w', or 'a'
        endpoint
            Server endpoint.   Defaults to "http://localhost:5000"
        username
            username for authentication
        password
            password for authentication
        bucket
            bucket (or storage container) to use for domain.  If not set, server default bucket will be used
        api_key
            user's api key (for server configurations that use api_key rather than username/password)
        logger
            supply log handler to be used
        owner
            set the owner to be used when new domain is created (defaults to username).  Only valid when used
            by admin users
        linked_domain
            Create new domain using the root of the linked domain
        retries
            Number of retry attempts to be used if a server request fails

        """

        self.log = logging.getLogger("h5pyd")

        if len(domain_name) == 0:
            raise ValueError("Invalid folder name")

        # strip prefix is any
        for protocol in ("http://", "https://", "hdf5://"):
            if domain_name.startswith(protocol):
                n = len(protocol) - 1
                domain_name = domain_name[n:]

        if domain_name[0] != "/":
            raise ValueError("Folder name must start with '/'")

        if domain_name[-1] != "/":
            raise ValueError("Folder name must end with '/'")

        if mode and mode not in ("r", "r+", "w", "w-", "x", "a"):
            raise ValueError("Invalid mode; must be one of r, r+, w, w-, x, a")

        self._pattern = pattern
        self._query = query

        if mode is None:
            mode = "r"

        cfg = Config()  # pulls in state from a .hscfg file (if found).

        if endpoint is None and "hs_endpoint" in cfg:
            endpoint = cfg["hs_endpoint"]

        if username is None and "hs_username" in cfg:
            username = cfg["hs_username"]

        if password is None and "hs_password" in cfg:
            password = cfg["hs_password"]

        if api_key is None and "hs_api_key" in cfg:
            api_key = cfg["hs_api_key"]

        if bucket is None and "hs_bucket" in cfg:
            bucket = cfg["hs_bucket"]

        if len(domain_name) <= 1:
            self._domain = None
        else:
            self._domain = domain_name[:-1]
        self._subdomains = None
        self._subdomain_marker = None
        self._batch_size = batch_size
        self._verbose = verbose

        self._http_conn = HttpConn(
            self._domain,
            endpoint=endpoint,
            username=username,
            password=password,
            bucket=bucket,
            api_key=api_key,
            mode=mode,
            logger=logger,
            retries=retries,
        )
        self.log = self._http_conn.logging

        domain_json = None

        # try to do a GET from the domain
        if domain_name == "/":
            req = "/domains"
        else:
            req = "/"

        # need some special logic for the first request in local mode
        # to give the sockets time to initialize
        if endpoint.startswith("local"):
            connect_backoff = [0.5, 1, 2, 4, 8, 16]
        else:
            connect_backoff = []

        connect_try = 0

        while True:
            try:
                rsp = self._http_conn.GET(req)
                break
            except IOError:
                if connect_try < len(connect_backoff):
                    time.sleep(connect_backoff[connect_try])
                else:
                    raise
                connect_try += 1

        if rsp.status_code in (404, 410) and mode in ("w", "w-", "x"):
            # create folder
            body = {"folder": True}
            if owner:
                body["owner"] = owner
            rsp = self._http_conn.PUT(req, body=body)
            if rsp.status_code != 201:
                self._http_conn.close()
                raise IOError(rsp.status_code, rsp.reason)
        elif rsp.status_code != 200:
            # folder must exist
            if rsp.status_code < 500:
                self.log.warning("status_code: {}".format(rsp.status_code))
            else:
                self.log.error("status_code: {}".format(rsp.status_code))
            raise IOError(rsp.status_code, rsp.reason)
        domain_json = json.loads(rsp.text)
        self.log.info("domain_json: {}".format(domain_json))
        if "class" in domain_json:
            if domain_json["class"] != "folder":
                self.log.warning("Not a folder domain")
            self._obj_class = domain_json["class"]
        elif "root" in domain_json:
            # open with Folder but actually has a root group
            self._obj_class = "domain"
        else:
            self._obj_class = "folder"
        self._name = domain_name
        if "created" in domain_json:
            self._created = domain_json["created"]
        else:
            self._created = None
        if "lastModified" in domain_json:
            self._modified = domain_json["lastModified"]
        else:
            self._modified = None
        if "owner" in domain_json:
            self._owner = domain_json["owner"]
        else:
            self._owner = None

    def getACL(self, username):
        if self._http_conn is None:
            raise IOError(400, "folder is not open")
        req = "/acls/" + username
        rsp = self._http_conn.GET(req)
        if rsp.status_code != 200:
            raise IOError(rsp.reason)
        rsp_json = json.loads(rsp.text)
        acl_json = rsp_json["acl"]
        return acl_json

    def getACLs(self):
        if self._http_conn is None:
            raise IOError(400, "folder is not open")
        req = "/acls"
        rsp = self._http_conn.GET(req)
        if rsp.status_code != 200:
            raise IOError(rsp.status_code, rsp.reason)
        rsp_json = json.loads(rsp.text)
        acls_json = rsp_json["acls"]
        return acls_json

    def putACL(self, acl):
        if self._http_conn is None:
            raise IOError(400, "folder is not open")
        if self._http_conn.mode == "r":
            raise IOError(400, "folder is open as read-onnly")
        if "userName" not in acl:
            raise IOError(404, "ACL has no 'userName' key")
        perm = {}
        for k in ("create", "read", "update", "delete", "readACL", "updateACL"):
            if k not in acl:
                raise IOError(404, "Missing ACL field: {}".format(k))
            perm[k] = acl[k]

        req = "/acls/" + acl["userName"]
        rsp = self._http_conn.PUT(req, body=perm)
        if rsp.status_code != 201:
            raise IOError(rsp.status_code, rsp.reason)

    def _getSubdomains(self):
        if self._http_conn is None:
            raise IOError(400, "folder is not open")
        if self._subdomains is not None and not self._subdomain_marker:
            # we've got all the subdomains, return 0 to indicate none were fetched
            return 0
        req = "/domains"
        if self._domain is None:
            params = {"domain": "/"}
        else:
            params = {"domain": self._domain + "/"}
        if self._verbose:
            params["verbose"] = 1  # to get lastModified
        if not self._query:
            params["Limit"] = self._batch_size  # get 100 at a time
        if self._pattern:
            params["pattern"] = self._pattern
        if self._query:
            params["query"] = self._query
        if self._subdomain_marker:
            params["Marker"] = self._subdomain_marker
        rsp = self._http_conn.GET(req, params=params)
        if rsp.status_code != 200:
            raise IOError(rsp.status_code, rsp.reason)
        rsp_json = json.loads(rsp.text)
        if "domains" not in rsp_json:
            raise IOError(500, "Unexpected Error")
        domains = rsp_json["domains"]
        count = len(domains)
        if self._subdomains is None:
            # setting to an empty list signifies we've done at least one request
            self._subdomains = []
        # append to what we have
        self._subdomains.extend(domains)
        if len(domains) == self._batch_size:
            # save the marker for the next batch
            self._subdomain_marker = domains[-1]["name"]
        else:
            self._subdomain_marker = None  # we got all the domains
        return count

    def close(self):
        """Clears reference to remote resource."""
        self._domain = None
        self._http_conn.close()
        self._http_conn = None

    def __getitem__(self, name):
        """Get a domain"""
        if self._http_conn is None:
            raise IOError(400, "folder is not open")
        if self._subdomains is None:
            self._getSubdomains()
        while True:
            domains = self._subdomains
            for domain in domains:
                if op.basename(domain["name"]) == name:
                    return domain
            # see if we can fetch more domains
            count = self._getSubdomains()
            if count == 0:
                break
        return None

    def delete_item(self, name, keep_root=False):
        """delete domain"""
        if self._http_conn is None:
            raise IOError(400, "folder is not open")
        if self._http_conn.mode == "r":
            raise IOError(400, "folder is open as read-onnly")
        if self._domain:
            domain = self._domain + "/" + name
        else:
            domain = "/" + name  # top-level delete
        headers = self._http_conn.getHeaders()
        req = "/"
        params = {"domain": domain}
        if keep_root:
            params["keep_root"] = 1
        rsp = self._http_conn.DELETE(req, headers=headers, params=params)
        if rsp.status_code != 200:
            raise IOError(rsp.status_code, rsp.reason)
        self._subdomains = None  # reset the cache list
        self._subdomain_marker = None

    def __delitem__(self, name):
        """Delete domain."""
        self.delete_item(name)

    def __len__(self):
        """Number of subdomains of this folder"""
        if self._http_conn is None:
            raise IOError(400, "folder is not open")
        count = 1
        while count > 0:
            # keep grabbing subdomains till there are no more to fetch
            count = self._getSubdomains()
        return len(self._subdomains)

    def __iter__(self):
        """Iterate over subdomain names"""
        if self._http_conn is None:
            raise IOError(400, "folder is not open")
        self._getSubdomains()
        index = 0
        while True:
            if len(self._subdomains) == index:
                # get more if we can
                count = self._getSubdomains()
                if count == 0:
                    break
            domain = self._subdomains[index]
            index += 1
            yield op.basename(domain["name"])

    def __contains__(self, name):
        """Test if a member name exists"""
        if self._http_conn is None:
            raise IOError(400, "folder is not open")
        self._getSubdomains()
        index = 0
        found = False
        while True:
            if len(self._subdomains) == index:
                # get more if we can
                count = self._getSubdomains()
                if count == 0:
                    break
            domain = self._subdomains[index]
            index += 1
            if op.basename(domain["name"]) == name:
                found = True
                break
        return found

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass

    def __repr__(self):

        return self.domain




import sys
import os.path as op
import logging
from datetime import datetime
import h5pyd as h5py
import numpy as np
if __name__ == "__main__":
    from config import Config
else:
    from .config import Config

#
# Print objects in a domain in the style of the hsls utilitiy
#

cfg = Config()

def intToStr(n):
    if cfg["human_readable"]:
        s = "{:,}".format(n)
    else:
        s = "{}".format(n)
    return s


def format_size(n):
    if n is None or n == ' ':
        return ' ' * 8
    symbol = ' '
    if not cfg["human_readable"]:
        return str(n)
    # convert to common storage unit
    for s in ('B', 'K', 'M', 'G', 'T'):
        if n < 1024:
            symbol = s
            break
        n /= 1024
    if symbol == 'B':
        return "{:7}B".format(n)
    else:
        return "{:7.1f}{}".format(n, symbol)


def getShapeText(dset):
    shape_text = "{NULL}"
    shape = dset.shape
    if shape is not None:
        shape_text = "{"
        rank = len(shape)
        if rank == 0:
            shape_text += "{SCALAR}"
        else:
            for dim in range(rank):
                if dim != 0:
                    shape_text += ", "
                shape_text += str(shape[dim])
                if cfg["verbose"]:
                    # get unlimited dimension
                    max_extent = dset.maxshape[dim]
                    shape_text += '/'
                    if max_extent is None:
                        shape_text += "Inf"
                    else:
                        shape_text += str(max_extent)
        shape_text += "}"
    return shape_text


def visititems(name, grp, visited):
    for k in grp:
        item = grp.get(k, getlink=True)
        class_name = item.__class__.__name__
        item_name = op.join(name, k)
        if class_name == "HardLink":
            # follow hardlinks
            try:
                item = grp.get(k)
                dump(item_name, item, visited=visited)
            except IOError:
                # object deleted but hardlink left?
                desc = "{Missing hardlink object}"
                print("{0:24} {1} {2}".format(item_name, class_name, desc))

        elif class_name == "SoftLink":
            desc = '{' + item.path + '}'
            print("{0:24} {1} {2}".format(item_name, class_name, desc))
        elif class_name == "ExternalLink":
            desc = '{' + item.path + '//' + item.filename + '}'
            print("{0:24} {1} {2}".format(item_name, class_name, desc))
        else:
            desc = '{Unknown Link Type}'
            print("{0:24} {1} {2}".format(item_name, class_name, desc))


def getTypeStr(dt):
    # TBD: convert dt to more HDF5-like representation
    return str(dt)


def dump(name, obj, visited=None):
    class_name = obj.__class__.__name__
    desc = None
    obj_id = None
    if class_name in ("Dataset", "Group", "Datatype", "Table"):
        obj_id = obj.id.id
        if visited and obj_id in visited:
            same_as = visited[obj_id]
            print("{0:24} {1}, same as {2}".format(name, class_name, same_as))
            return
    elif class_name in ("ExternalLink", "SoftLink"):
        pass
    else:
        raise TypeError(f"unexpected classname: {class_name}")

    is_dataset = False
    if class_name in ("Dataset", "Table"):
        is_dataset = True


    if is_dataset:
        desc = getShapeText(obj)
        obj_id = obj.id.id
    elif class_name == "Group":
        obj_id = obj.id.id
    elif class_name == "Datatype":
        obj_id = obj.id.id
    elif class_name == "SoftLink":
        desc = '{' + obj.path + '}'
    elif class_name == "ExternalLink":
        desc = '{' + obj.filename + '//' + obj.path + '}'

    if desc is None:
        print("{0} {1}".format(name, class_name))
    else:
        print("{0} {1} {2}".format(name, class_name, desc))

    if cfg["verbose"] and obj_id is not None:
        print("    {0:>32}: {1}".format("UUID", obj_id))

    if cfg["verbose"] and is_dataset and obj.shape is not None \
            and obj.chunks is not None:
        chunk_size = obj.dtype.itemsize
        
        if isinstance(obj.id.layout, dict):
            # H5D_CHUNKED_REF layout
            chunk_dims = obj.id.layout["dims"]
            obj_layout = obj.id.layout["class"]
        else:
            chunk_dims = obj.chunks
            obj_layout = "H5D_CHUNKED"
        storage_desc = f"Storage {obj_layout}"
        max_chunk_count = 1
        rank = len(obj.shape)
        for i in range(rank):
            extent = obj.shape[i]
            chunk_dim = chunk_dims[i]
            chunk_size *= chunk_dim
            max_chunk_count *= -(-extent // chunk_dim)
        dset_size = obj.dtype.itemsize
        for dim_extent in obj.shape:
            dset_size *= dim_extent

        if obj_layout == "H5D_CHUNKED_REF_INDIRECT":
            chunk_table_id = obj.id.layout["chunk_table"]
            chunk_table = obj.file[f"datasets/{chunk_table_id}"]
            num_chunks = int(np.prod(chunk_table.shape))
            chunk_table_elements = chunk_table[...].reshape((num_chunks,))
            num_linked_chunks = 0
            allocated_size = 0
            for e in chunk_table_elements:
                chunk_offset = e[0]
                chunk_size = e[1]
                if chunk_offset > 0 and chunk_size > 0:
                    num_linked_chunks += 1
                    allocated_size += chunk_size
            num_chunks = num_linked_chunks
            chunk_type = "linked"

        else:
            num_chunks = obj.num_chunks
            allocated_size = obj.allocated_size
            chunk_type = "allocated"

        if num_chunks is not None and allocated_size is not None:
            fstr = "    {0:>32}: {1} {2} bytes, {3}/{4} {5} chunks"

            s = fstr.format("Chunks", chunk_dims, intToStr(chunk_size), intToStr(num_chunks), 
                            intToStr(max_chunk_count), chunk_type)
            print(s)
            if dset_size > 0:
                utilization = allocated_size / dset_size
                fstr = "    {0:>32}: {1} logical bytes, {2} {3} bytes, {4:.2f}% utilization"
                print(fstr.format(storage_desc, intToStr(dset_size),
                                  intToStr(allocated_size),
                                  chunk_type,
                                  utilization * 100.0))
            else:
                fstr = "    {0:>32}: {1} logical bytes, {2} {3} bytes"
                print(fstr.format(storage_desc, intToStr(dset_size),
                                  intToStr(allocated_size), chunk_type))

        else:
            # verbose info not available, just show the chunk layout
            fstr = "    {0:>32}: {1} {2} bytes"
            print(fstr.format("Chunks", chunk_dims, intToStr(chunk_size)))

        # show filters (if any)
        # currently HSDS only supports the shuffle filter (not fletcher32 or 
        # scaleoffset), so just check for shuffle and whatever compressor may
        # be applied
        filter_number = 0
        fstr = "    {0:>30}-{1}: {2} OPT {{{3}}}"
        if obj.compression_opts:
            compression_opts = obj.compression_opts
        else:
            compression_opts = ""
        if obj.shuffle:
            print(fstr.format("Filter", filter_number, "shuffle", compression_opts))
            filter_number += 1
        if obj.compression:
            print(fstr.format("Filter", filter_number, obj.compression, compression_opts))

        # display type info

        fstr = "    {0:>32}: {1}"
        print(fstr.format("Type", getTypeStr(obj.dtype)))  # dump out type info

    if cfg["showattrs"] and class_name in ("Dataset", "Table", "Group", "Datatype"):
        # dump attributes for the object
        for attr_name in obj.attrs:
            attr = obj.attrs[attr_name]
            el = "..."  # show this if the attribute is too large
            if isinstance(attr, np.ndarray):
                rank = len(attr.shape)
            else:
                rank = 0  # scalar data
            if rank > 1:
                val = "[" * rank + el + "]" * rank
                print("   attr: {0:24} {1}".format(attr_name, val))
            elif rank == 1 and attr.shape[0] > 1:
                val = "[{},{}]".format(attr[0], el)
                print("   attr: {0:24} {1}".format(attr_name, val))
            else:
                print("   attr: {0:24} {1}".format(attr_name, attr))

    if visited is not None and obj_id is not None:
        visited[obj_id] = name
    if class_name == "Group" and visited is not None:
        visititems(name, obj, visited)


def dumpACL(acl):
    perms = ""
    if acl["create"]:
        perms += 'c'
    else:
        perms += '-'
    if acl["read"]:
        perms += 'r'
    else:
        perms += '-'
    if acl["update"]:
        perms += 'u'
    else:
        perms += '-'
    if acl["delete"]:
        perms += 'd'
    else:
        perms += '-'
    if acl["readACL"]:
        perms += 'e'
    else:
        perms += '-'
    if acl["updateACL"]:
        perms += 'p'
    else:
        perms += '-'
    print("    acl: {0:24} {1}".format(acl["userName"], perms))


def dumpAcls(obj):
    try:
        acls = obj.getACLs()
    except IOError:
        print("read ACLs is not permitted")
        return

    for acl in acls:
        dumpACL(acl)


def getFolder(domain):
    username = cfg["hs_username"]
    password = cfg["hs_password"]
    endpoint = cfg["hs_endpoint"]
    bucket   = cfg["hs_bucket"]
    pattern = cfg["pattern"]
    query = cfg["query"]
    if cfg["verbose"]:
        verbose = True
    else:
        verbose = False
    batch_size = 100  # use smaller batchsize for interactively listing of large collections
    d = h5py.Folder(domain, endpoint=endpoint, username=username, verbose=verbose,
                      password=password, bucket=bucket, pattern=pattern, query=query, batch_size=batch_size)
    return d


def getFile(domain):
    username = cfg["hs_username"]
    password = cfg["hs_password"]
    endpoint = cfg["hs_endpoint"]
    bucket = cfg["hs_bucket"]
    fh = h5py.File(domain, mode='r', endpoint=endpoint, username=username,
                   password=password, bucket=bucket, use_cache=True)
    return fh

def isFile(domain):
    username = cfg["hs_username"]
    password = cfg["hs_password"]
    endpoint = cfg["hs_endpoint"]
    bucket = cfg["hs_bucket"]

    return h5py.is_hdf5(domain, endpoint=endpoint, username=username,
                   password=password, bucket=bucket)



def visitDomains(domain, depth=1):
    if depth == 0:
        return 0

    count = 0
    if domain[-1] == '/':
        domain = domain[:-1]  # strip off trailing slash

    try:
        d = getFolder(domain + '/')
        dir_class = "domain"
        display_name = domain
        num_bytes = ' '
        if d.is_folder:
            dir_class = "folder"
            display_name += '/'
        elif cfg["verbose"]:
            # get the number of allocated bytes
            f = getFile(domain)
            num_bytes = f.total_size
            f.close()

        owner = d.owner
        if owner is None:
            owner = ""
        if d.modified is None:
            timestamp = ""
        else:
            timestamp = datetime.fromtimestamp(int(d.modified))

        print("{:35} {:15} {:8} {} {}".format(owner, format_size(num_bytes),
                                              dir_class, timestamp,
                                              display_name))
        count += 1
        if cfg["showacls"]:
            dumpAcls(d)
        for name in d:
            item = d[name]
            owner = item["owner"]
            full_path = domain + '/' + name

            num_bytes = " "
            if cfg["verbose"] and "total_size" in item:
                num_bytes = item["total_size"]
                if "total_size" not in cfg:
                    cfg["total_size"] = 0
                cfg["total_size"] += item["total_size"]
            else:
                 num_bytes = " "
            dir_class = item["class"]
            if item["lastModified"] is None:
                timestamp = ""
            else:
                timestamp = datetime.fromtimestamp(int(item["lastModified"]))

            print("{:35} {:15} {:8} {} {}".format(owner, format_size(num_bytes),
                                              dir_class, timestamp,
                                              full_path))
            if cfg["showacls"]:
                if dir_class == "folder":
                    with getFolder(domain + '/' + name + '/') as f:
                        dumpAcls(f)
                else:
                    with getFile(domain + '/' + name) as f:
                        dumpAcls(f)
            count += 1

            if dir_class == "folder":
                # recurse for folders
                n = visitDomains(domain + '/' + name, depth=(depth - 1))
                count += n

    except IOError as oe:
        if oe.errno in (403, 404, 410):
            # Ignore domains for which:
            #   * we don't have permsssions (403)
            #   * not found error (404)
            #   * recently deleted (410)
            pass
        else:
            print("error getting domain:", domain)
            sys.exit(str(oe))

    return count


def checkDomain(path):
    """ Convenience method to specify a domain + h5path as a single string.
        Walk up the path items, as soon as the parent is a domain or folder return it.
        Supply the other part as h5path.  """
    
    path_names = path.split("/")
    h5path = ""
    while path_names:
        domain_path = "/".join(path_names)
        if h5py.is_hdf5(domain_path):
            return (h5path, domain_path)
        last = path_names[-1]
        path_names = path_names[:-1]
        h5path = last + "/" + h5path

    return None

         




#
# Usage
#
def usage():
    option_names = cfg.get_names()
    cmd = cfg.get_cmd()
    print("Usage:\n")
    print(f"    {cmd} [ OPTIONS ]  domain")
    print(f"    {cmd} [ OPTIONS ]  folder")
    print("")
    print("Description:")
    print("    List contents of a domain or folder")
    print("       domain: HSDS domain (absolute path with or without 'hdf5:// prefix)")
    print("       folder: HSDS folder (path as above ending in '/')")
    print("")
    
    print("Options:")
    for name in option_names:
        help_msg = cfg.get_help_message(name)
        if help_msg:
            print(f"    {help_msg}")  
    print("")
    print(f"example: {cmd} -r -e http://hsdshdflab.hdfgroup.org /shared/tall.h5")
    print("")
    print(cfg.get_see_also(cmd))
    print("")
    sys.exit()


#
# Main
#
def main():
    domains = []

    # additional options
    cfg.setitem("showacls", False, flags=["--showacls",], help="display domain ACLs")
    cfg.setitem("showattrs", False, flags=["--showattrs",], help="display domain attributes")
    cfg.setitem("pattern", None, flags=["--pattern",], choices=["REGEX",], help="list domains that match the given regex")
    cfg.setitem("query", None, flags=["--query",], choices=["QUERY",], help="list domains where the attributes of the root group match the given query string")
    cfg.setitem("recursive", False, flags=["-r", "--recursive"], help="recursively list sub-folders or sub-groups")
    cfg.setitem("dataset_path", None, flags=["-d", "--dataset"], choices=["H5PATH",], help="display specified dataset")
    cfg.setitem("group_path", None, flags=["-g", "--group"], choices=["H5PATH",], help="display specified group")
    cfg.setitem("datatype_path", None, flags=["-t", "--datatype"], choices=["H5PATH",], help="display specified datatype")
    cfg.setitem("human_readable", False, flags=["-H", "--human-readable"], help="with -v, print human readable sizes (e.g. 123M)")
    cfg.setitem("help", False, flags=["-h", "--help"], help="this message")

    try:
        domains = cfg.set_cmd_flags(sys.argv[1:])
    except ValueError as ve:
        print(ve)
        usage()

    if len(domains) == 0:
        # need a domain - use root
        domains.append("/")

    # setup logging
    logfname = cfg["logfile"]
    loglevel = cfg.get_loglevel()
    logging.basicConfig(filename=logfname, format='%(levelname)s %(asctime)s %(message)s', level=loglevel)
    logging.debug(f"set log_level to {loglevel}")

    for domain in domains:
        if cfg["recursive"]:
            depth = -1
        else:
            depth = 1

        h5path = None

        if domain.endswith('/'):
            # given a folder path
            count = visitDomains(domain, depth=depth)
            print(f"{count} items")

        else:
            res = checkDomain(domain)
            if res is None:
                # couldn't find a domain, call getFile anyway so we can 
                # report on exactly what went wrong
                pass
            else:
                h5path = res[0]
                domain = res[1]
                logging.debug(f"using h5path: {h5path} domain: {domain}")
            try:
                f = getFile(domain)
            except IOError as ioe:
                if ioe.errno == 401:
                    print("Username/Password missing or invalid")
                    continue
                if ioe.errno == 403:
                    print(f"No permission to read domain: {domain}")
                    continue
                elif ioe.errno == 404:
                    print(f"Domain {domain} not found")
                    continue
                elif ioe.errno == 410:
                    print(f"Domain {domain} has been removed")
                    continue
                else:
                    print(f"Unexpected error: {ioe}")
                    continue

            grp = f['/']
            if grp is None:
                print(f"{domain}: No such domain")
                domain += '/'
                count = visitDomains(domain, depth=depth)
                print(f"{count} items")
                continue

            if h5path:
                if h5path not in f:
                    print(f"h5path: {h5path} not found in domain: {domain}")
                    continue

                obj = f[h5path]
                class_name = obj.__class__.__name__
                if class_name in ("Table", "Dataset"):
                    dump(h5path, obj)
                    continue
                elif class_name == "Datatype":
                    dump(h5path, obj)
                    continue
                else:
                    grp = obj
                    # we'll just fall through to our normal processin

            if cfg["group_path"]:
                if h5path:
                    print("group_path option can't be used when h5path is specified")
                    continue
                h5path = cfg["group_path"]
                if h5path not in grp:
                    print(f"group_path: {h5path} not found")
                    continue
                else:
                    grp = grp[h5path]
                    class_name = grp.__class__.__name__
                    if class_name != "Group":
                        if class_name in ("Table", "Dataset"):
                            hint = " (use --dataset option to display)"
                        elif class_name == "Datatype":
                            hint = " (use --datatype option to display)"
                        else:
                            hint = ""
                        print(f"group_path: {h5path} points to a {class_name}{hint}")
                        continue
            else:
                h5path = "/"  # start at root

            if cfg["dataset_path"]:
                dataset_path = cfg["dataset_path"]
                if dataset_path[0] == "/" and h5path != "/":
                    print("--group_path can't be used with absolute --dataset_path")
                    continue
                if dataset_path[0] == "/":
                    h5path = dataset_path  # replace h5path
                else:
                    if h5path[-1] != "/":
                        h5path = h5path + "/"
                    h5path = h5path + dataset_path
                    print("using h5path:", h5path) 
                if h5path not in grp:
                    print("dataset path: {h5path} not found")
                    continue
                obj = grp[h5path]
                class_name = obj.__class__.__name__
                if class_name not in ("Table", "Dataset"):
                    print(f"was expecting a Dataset object but found: {class_name}")
                    continue

                dump(h5path, obj)
                continue

            if cfg["datatype_path"]:
                datatype_path = cfg["datatype_path"]
                if datatype_path[0] == "/" and h5path != "/":
                    print("--group_path can't be used with absolute --datatype_path")
                    continue
                if datatype_path[0] == "/":
                    h5path = datatype_path  # replace h5path
                else:
                    if h5path[-1] != "/":
                        h5path = h5path + "/"
                    h5path = h5path + datatype_path
                    print("using h5path:", h5path) 
                if h5path not in grp:
                    print("datatype path: {h5path} not found")
                    continue
                obj = grp[h5path]
                class_name = obj.__class__.__name__
                if class_name != "Datatype":
                    print(f"was expecting a Datatype object but found: {class_name}")
                    continue
                dump(h5path, obj)
                continue

            if depth < 0:
                # recursive
                visited = {}  # dict of id to h5path
                visited[grp.id.id] = '/'
                visititems('/', grp, visited)
            else:
                for k in grp:
                    item = grp.get(k, getlink=True)
                    if item.__class__.__name__ == "HardLink":
                        # follow hardlinks
                        try:
                            item = grp[k]
                        except IOError:
                            # object deleted?  Just dump link info
                            pass
                    dump(k, item)
            if cfg["showacls"]:
                dumpAcls(grp)
            grp.file.close()
    if "total_size" in cfg and cfg["total_size"] > 0:
        # print aggregate size
        total_size = cfg["total_size"]
        s = format_size(total_size)
        s = s.strip()

        print(f"{s} bytes")


if __name__ == "__main__":
    main()



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

from __future__ import absolute_import
import numpy
from .base import  _decode   
from .base import bytesToArray
from .dataset import Dataset
from .objectid import DatasetID
from . import selections as sel
from .h5type import Reference
from .h5type import check_dtype
from .h5type import getQueryDtype   


class Cursor():
    """
      Cursor for retreiving rows from a table
      buffer_rows can be used to control how many rows
      will be fetched from the server
    """
    def __init__(self, table, query=None, start=None, stop=None, buffer_rows=None):
        self._table = table
        self._query = query
        DEFAULT_BUFFER_BYTES = 1000000
        if buffer_rows is None:
            buffer_rows = DEFAULT_BUFFER_BYTES // table.dtype.itemsize
        if buffer_rows < 1:
            buffer_rows = 1
        self._buffer_rows = buffer_rows

        if start is None:
            self._start = 0
        else:
            self._start = start
        if stop is None:
            self._stop = table.nrows
        else:
            self._stop = stop

    def __iter__(self):
        """ Iterate over the first axis.  TypeError if scalar.

        BEWARE: Modifications to the yielded data are *NOT* written to file.
        """
        nrows = self._stop - self._start

        arr = None
        query_complete = False

        for indx in range(self._stop - self._start):
            if indx % self._buffer_rows == 0:
                # grab another buffer
                read_count = self._buffer_rows
                if nrows - indx < read_count:
                    read_count = nrows - indx
                if self._query is None:
                    arr = self._table[indx+self._start:read_count+indx+self._start]
                else:
                    # call table to return query result
                    if query_complete:
                        arr = None  # nothing more to fetch
                    else:
                        arr = self._table.read_where(self._query, start=indx+self._start, limit=read_count)
                        if arr is not None and arr.shape[0] < read_count:
                            query_complete = True  # we've gotten all the rows
            if arr is not None and indx%self._buffer_rows < arr.shape[0]:
                yield arr[indx%self._buffer_rows]

class Table(Dataset):

    """
        Represents an HDF5 dataset
    """
    def __init__(self, bind):
        """ Create a new Table object by binding to a low-level DatasetID.
        """

        if not isinstance(bind, DatasetID):
            raise ValueError("%s is not a DatasetID" % bind)
        Dataset.__init__(self, bind)

        if len(self._dtype) < 1:
            raise ValueError("Table type must be compound")

        if len(self._shape) > 1:
            raise ValueError("Table must be one-dimensional")


    @property
    def colnames(self):
        """Numpy-style attribute giving the number of dimensions"""
        names = []
        for field in self._dtype.descr:
            # each element should be a tuple ('fieldname', dt)
            names.append(field[0])
        return names

    @property
    def nrows(self):
        return self._shape[0]

    def read(self, start=None, stop=None, step=None, field=None, out=None):
        if start is None:
            start = 0
        if stop is None:
            stop = self._shape[0]
        if step is None:
            step = 1
        arr = self[start:stop:step]
        if field is not None:
            #TBD - read just the field once the service supports it
            tmp = arr[field]
            arr = tmp
        if out is not None:
            # TBD - read direct
            numpy.copyto(out, arr)
        else:
            return arr



    def read_where(self, condition, condvars=None, field=None, start=None, stop=None, step=None, limit=0, include_index=True):
        """Read rows from table using pytable-style condition
        """
        names = ()  # todo
        def readtime_dtype(basetype, names):
            """ Make a NumPy dtype appropriate for reading """

            if len(names) == 0:  # Not compound, or we want all fields
                return basetype

            if basetype.names is None:  # Names provided, but not compound
                raise ValueError("Field names only allowed for compound types")

            for name in names:  # Check all names are legal
                if not name in basetype.names:
                    raise ValueError("Field %s does not appear in this type." % name)

            return numpy.dtype([(name, basetype.fields[name][0]) for name in names])

        new_dtype = getattr(self._local, 'astype', None)
        if new_dtype is not None:
            mtype = readtime_dtype(new_dtype, names)
        else:
            # This is necessary because in the case of array types, NumPy
            # discards the array information at the top level.
            mtype = readtime_dtype(self.dtype, names)
        # todo - will need the following once we have binary transfers
        # mtype = h5t.py_create(new_dtype)
        rsp_type = getQueryDtype(mtype)
        

        # Perform the dataspace selection
        if start or stop:
            if not start:
                start = 0
            if not stop:
                stop = self._shape[0]
        else:
            start = 0
            stop = self._shape[0]

        selection_arg = slice(start, stop)
        selection = sel.select(self, selection_arg)

        if selection.nselect == 0:
            return numpy.ndarray(selection.mshape, dtype=mtype)

        # setup for pagination in case we can't read everthing in one go
        data = []  # one ndarray for each request response
        cursor = start
        page_size = stop - start
        total_rows = 0

        while True:
            # Perfom the actual read
            req = "/datasets/" + self.id.uuid + "/value"
            params = {}
            params["query"] = condition
            if limit > 0:
                params["Limit"] = limit - total_rows
            self.log.info("req - cursor: {} page_size: {}".format(cursor, page_size))
            end_row = cursor+page_size
            if end_row > stop:
                end_row = stop
            selection_arg = slice(cursor, end_row)
            selection = sel.select(self, selection_arg)

            sel_param = selection.getQueryParam()
            self.log.debug("query param: {}".format(sel_param))
            if sel_param:
                params["select"] = sel_param
            try:
                self.log.debug("params: {}".format(params))
                rsp = self.GET(req, params=params)
                if isinstance(rsp, bytes):
                    # binary response
                    arr = bytesToArray(rsp, rsp_type, None) 
                    count = len(arr)   
                    self.log.info(f"got {count} rows binary data")
                else:
                    values = rsp["value"]
                    count = len(values)
                    if "index" in rsp:
                        # older server version that returns index as a seperate key
                        indices = rsp["index"]
                        if len(indices) != count:
                            raise ValueError(f"expected {count} indicies, but got: {len(indices)}")
                    else:
                        indices = None
                    count = len(values)
                    self.log.info(f"got {count} rows json data")
                    # convert to numpy array
                    arr = numpy.empty((count,), dtype=rsp_type)
                    for i in range(count):
                        if indices is not None:
                            e = [indices[i],]
                            e.extend(values[i])
                        else:
                            e = values[i]
                        arr[i] = tuple(e)
                        
                    self.log.info("got {} rows".format(count))
                total_rows += count
                data.append(arr)

                # advance to next page
                cursor += page_size
            except IOError as ioe:
                if ioe.errno == 413 and page_size > 1024:
                    # too large a query target, try reducing the page size
                    # if it is not already relatively small (1024)
                    page_size //= 2
                    page_size += 1  # bump up to avoid tiny pages in the last iteration
                    self.log.info("Got 413, reducing page_size to: {}".format(page_size))
                else:
                    # otherwise, just raise the exception
                    self.log.info("Unexpected exception: {}".format(ioe.errno))
                    raise ioe
            if cursor >= stop or (limit > 0 and total_rows == limit):
                self.log.info("completed iteration, returning: {} rows".format(len(data)))
                break

        # need some special conversion for compound types --
        # each element must be a tuple, but the JSON decoder
        # gives us a list instead.

        if len(data) == 0:
            raise ValueError("unexpected list size")
        # combine arrays
        if len(data) > 1:
            ret_arr = numpy.empty((total_rows,), dtype=rsp_type)
            start = 0
            for arr in data:
                nrows = len(arr)
                ret_arr[start:(start+nrows)] = arr[:]
                start += nrows
        else:
            ret_arr = data[0]

        return ret_arr


    def update_where(self, condition, value, start=None, stop=None, step=None, limit=None):
        """Modify rows in table using pytable-style condition
        """
        if not isinstance(value, dict):
            raise ValueError("expected value to be a dict")

        # Perform the dataspace selection
        if start or stop:
            if not start:
                start = 0
            if not stop:
                stop = self._shape[0]
        else:
            start = 0
            stop = self._shape[0]

        selection_arg = slice(start, stop)
        selection = sel.select(self, selection_arg)
        sel_param = selection.getQueryParam()
        params = {}
        params["query"] = condition
        if limit:
            params["Limit"] = limit
        self.log.debug(f"query param: {sel_param}")
        if sel_param:
            params["select"] = sel_param

        req = "/datasets/" + self.id.uuid + "/value"

        rsp = self.PUT(req, body=value, format="json", params=params)
        indices = None
        arr = None
        if "index" in rsp:
            indices = rsp["index"]
        elif "value" in rsp:
            # new-style return type - index is first element in each row
            indices = []
            for row in rsp["value"]:
                indices.append(row[0])
        else:
            raise ValueError("unexpected response from PUT query")
        if indices:
            arr = numpy.array(indices)

        return arr

    def create_cursor(self, condition=None,  start=None, stop=None):
        """Return a cursor for iteration
        """
        return Cursor(self, query=condition, start=start, stop=stop)



    def append(self, rows):
        """ Append rows to end of table
        """
        self.log.info("Table append")
        if not self.id.uuid.startswith("d-"):
            # Append ops only work with HSDS
            raise ValueError("append not supported")

        if self._item_size != "H5T_VARIABLE":
            use_base64 = True   # may need to set this to false below for some types
        else:
            use_base64 = False  # never use for variable length types
            self.log.debug("Using JSON since type is variable length")

        val = rows  # for compatibility with dataset code...
        # get the val dtype if we're passed a numpy array
        val_dtype = None
        try:
            val_dtype = val.dtype
        except AttributeError:
            pass # not a numpy object, just leave dtype as None

        if isinstance(val, Reference):
            # h5pyd References are just strings
            val = val.tolist()

        # Generally we try to avoid converting the arrays on the Python
        # side.  However, for compound literals this is unavoidable.
        # For h5pyd, do extra check and convert type on client side for efficiency
        vlen = check_dtype(vlen=self.dtype)
        if vlen is not None and vlen not in (bytes, str):
            self.log.debug("converting ndarray for vlen data")
            try:
                val = numpy.asarray(val, dtype=vlen)
            except ValueError:
                try:
                    val = numpy.array([numpy.array(x, dtype=vlen)
                                       for x in val], dtype=self.dtype)
                except ValueError:
                    pass
            if vlen == val_dtype:
                if val.ndim > 1:
                    tmp = numpy.empty(shape=val.shape[:-1], dtype=object)
                    tmp.ravel()[:] = [i for i in val.reshape(
                        (numpy.product(val.shape[:-1]), val.shape[-1]))]
                else:
                    tmp = numpy.array([None], dtype=object)
                    tmp[0] = val
                val = tmp

        elif isinstance(val, numpy.ndarray):
            # convert array if needed
            # TBD - need to handle cases where the type shape is different
            self.log.debug("got numpy array")
            if val.dtype != self.dtype and val.dtype.shape == self.dtype.shape:
                self.log.info("converting {} to {}".format(val.dtype, self.dtype))
                # convert array
                tmp = numpy.empty(val.shape, dtype=self.dtype)
                tmp[...] = val[...]
                val = tmp
        else:
            val = numpy.asarray(val, order='C', dtype=self.dtype)

        self.log.debug("rows shape: {}".format(val.shape))
        self.log.debug("data dtype: {}".format(val.dtype))

        if len(val.shape) != 1:
            raise ValueError("rows must be one-dimensional")

        numrows = val.shape[0]

        req = "/datasets/" + self.id.uuid + "/value"

        params = {}
        body = {}


        format = "json"

        if use_base64:

            # server is HSDS, use binary data, use param values for selection
            format = "binary"
            body = val.tobytes()
            self.log.debug("writing binary data, {} bytes".format(len(body)))
            params["append"] = numrows
        else:
            if type(val) is not list:
                val = val.tolist()
            val = _decode(val)
            self.log.debug("writing json data, {} elements".format(len(val)))
            self.log.debug("data: {}".format(val))
            body['value'] = val
            body['append'] = numrows

        self.PUT(req, body=body, format=format, params=params)

        # if we get here, the request was successful, adjust the shape
        total_rows = self._shape[0] + numrows
        self._shape = (total_rows,)



import sys
import os.path as op
import logging
import h5pyd as h5py
if __name__ == "__main__":
    from config import Config
else:
    from .config import Config

#
# Create domain or update timestamp if domain already exist
#

cfg = Config()

def getFolder(domain):
    username = cfg["hs_username"]
    password = cfg["hs_password"]
    endpoint = cfg["hs_endpoint"]
    bucket = cfg["hs_bucket"]
    logging.debug(f"getFolder({domain})")
    dir = h5py.Folder(domain, endpoint=endpoint, username=username, password=password, bucket=bucket)
    return dir

def createFolder(domain):
    username = cfg["hs_username"]
    password = cfg["hs_password"]
    endpoint = cfg["hs_endpoint"]
    bucket = cfg["hs_bucket"]
    owner = None
    if "hs_owner" in cfg:
        owner=cfg["hs_owner"]
    logging.debug(f"createFolder({domain})")
    dir = h5py.Folder(domain, mode='x', endpoint=endpoint, username=username, password=password, bucket=bucket, owner=owner)
    return dir

def getFile(domain, mode="a"):
    username = cfg["hs_username"]
    password = cfg["hs_password"]
    endpoint = cfg["hs_endpoint"]
    bucket = cfg["hs_bucket"]
    logging.debug(f"getFile(domain={domain}, mode={mode})")
    fh = h5py.File(domain, mode=mode, endpoint=endpoint, username=username, password=password, bucket=bucket)
    return fh

def createFile(domain):
    username = cfg["hs_username"]
    password = cfg["hs_password"]
    endpoint = cfg["hs_endpoint"]
    bucket = cfg["hs_bucket"]
    owner = None
    if "hs_owner" in cfg:
        owner=cfg["hs_owner"]
    logging.debug(f"createFile({domain})")
    fh = h5py.File(domain, mode='x', endpoint=endpoint, username=username, password=password, bucket=bucket, owner=owner)
    return fh

def getParentDomain(domain):
    if domain[-1] == '/':
        if len(domain) > 1:
            domain = domain[:-1]
    parent_domain = op.dirname(domain)
    if not parent_domain.endswith("/"):
        parent_domain += "/"
    return parent_domain

def touchDomain(domain):
    # get handle to parent folder
    parent_domain = getParentDomain(domain)

    if parent_domain == "/":
        if len(domain) < 4:
            msg = "Top-level folders must be at least three characters"
            logging.error(msg)
            sys.exit(msg)

    else:
        try:
            getFolder(parent_domain)
        except IOError as oe:
            #print("errno:", oe.errno)
            if oe.errno in (404, 410):   # Not Found
                sys.exit(f"Parent domain: {parent_domain} not found")
            elif oe.errno == 401:  # Unauthorized
                sys.exit("Authorization failure")
            elif oe.errno == 403:  # Forbidden
                sys.exit("Not allowed")
            else:
                sys.exit(f"Unexpected error: {oe}")

    hdomain = None
    try:
        if domain.endswith("/"):
            hdomain = getFolder(domain)
        else:
            hdomain = getFile(domain, mode="r")
    except IOError as oe:
        if oe.errno in (404, 410):   # Not Found
            pass  # domain not found
        else:
            sys.exit(f"Unexpected error: {oe}")

    if hdomain is not None:
        logging.debug(f"domain: {domain} exists")
        if domain.endswith("/"):
            sys.exit("Can not update timestamp of folder object")
        else:
            try:
                # get domain for updating
                hdomain = getFile(domain, mode="a")
                r = hdomain['/']
                # create/update attribute to update lastModified timestamp of domain
                r.attrs["hstouch"] = 1
                cfg.print(f"updated timestamp for domain: {domain}")
                hdomain.close()
            except IOError as oe:
                msg = f"Got error updating domain: {oe}"
                logging.error(msg)
                sys.exit(msg)
    else:
        # create domain
        if not domain.endswith("/"):
            try:
                fh = createFile(domain)
                cfg.print(f"domain created: {domain}, root id: {fh.id.id}")
                fh.close()
            except IOError as oe:
                sys.exit(f"Got error updating domain: {oe}")
        else:
            # make folder
            try:
                fh = createFolder(domain)
                cfg.print(f"folder created {domain}")
                fh.close()
            except IOError as oe:
                sys.exit(f"Got error updating domain: {oe}")

#
# Usage
#
def usage():
    option_names = cfg.get_names()
    cmd = cfg.get_cmd()
    print("Usage:\n")
    print(f"    {cmd} [ OPTIONS ]  domain")
    print(f"    {cmd} [ OPTIONS ]  folder")
    print("")
    print("Description:")
    print("    Create a new domain or folder")
    print("       domain: HSDS domain (absolute path with or without 'hdf5:// prefix)")
    print("       folder: HSDS folder (path as above ending in '/')")
    print("")
    print("Options:")
    for name in option_names:
        help_msg = cfg.get_help_message(name)
        if help_msg:
            print(f"    {help_msg}")  
    print("")
    print(f"Example: {cmd}  hdf5://home/myfolder/emptydomain.h5")
    print(cfg.get_see_also(cmd))
    print("")
    sys.exit()

#
# Main
#
def main():
    domains = []
    # additional options
    cfg.setitem("hs_owner", None, flags=["-o", "--owner"], choices=["OWNER",], help="set owner (must be run as an admin user)")
    cfg.setitem("help", False, flags=["-h", "--help"], help="this message")

    try:
        domains = cfg.set_cmd_flags(sys.argv[1:])
    except ValueError as ve:
        print(ve)
        usage()

    if len(domains) == 0:
        # need a domain
        usage()

    # setup logging
    logfname = cfg["logfile"]
    loglevel = cfg.get_loglevel()
    logging.basicConfig(filename=logfname, format='%(levelname)s %(asctime)s %(message)s', level=loglevel)
    logging.debug(f"set log_level to {loglevel}")

    for domain in domains:
        touchDomain(domain)

if __name__ == "__main__":
    main()



import sys
import os.path as op
import logging
import h5pyd as h5py
if __name__ == "__main__":
    from config import Config
else:
    from .config import Config

#
# Delete domain
#

cfg = Config()

def getFolder(domain, mode='r'):
    username = cfg["hs_username"]
    password = cfg["hs_password"]
    endpoint = cfg["hs_endpoint"]
    bucket = cfg["hs_bucket"]
    folder = h5py.Folder(domain, mode=mode, endpoint=endpoint, username=username, password=password, bucket=bucket)
    return folder

def exitUnlessIgnore(msg):
    if cfg["ignore"]:
        return
    sys.exit(msg)



def deleteDomain(domain):

    # get handle to parent folder
    if domain.endswith('/'):
        path = domain[:-1]
    else:
        path = domain
    parent_domain = op.dirname(path)
    base_name = op.basename(path)

    if len(parent_domain) < 2:
        #sys.exit("can't get parent domain")
        parent_domain = '/'

    if not parent_domain.endswith('/'):
        parent_domain += '/'
    try:
        hparent = getFolder(parent_domain, mode='a')
    except IOError as oe:
        if oe.errno == 404:   # Not Found
            msg = f"Parent domain: {parent_domain} not found"
            logging.error(msg)
            exitUnlessIgnore(msg)
        elif oe.errno == 401:  # Unauthorized
            msg = f"Authorization failure opening {parent_domain}"
            logging.error(msg)
            exitUnlessIgnore(msg)
        elif oe.errno == 403:  # Forbidden
            msg = f"Not allowed to open: {parent_domain}"
            logging.error(msg)
            exitUnlessIgnore(msg)
        else:
            msg = f"Unexpected error: {oe}"
            logging.error(msg)
            exitUnlessIgnore(msg)

    if base_name not in hparent:
        # note - this may happen if the domain was recently created and not
        # yet synced to S3
        msg = f"domain: {domain} not found"
        logging.error(msg)
        exitUnlessIgnore(msg)

    # delete the domain
    try:
        del hparent[base_name]
    except IOError as oe:
        if oe.errno == 404:   # Not Found
            # should have caught this in the base_name check...
            msg = f"domain {parent_domain} not found"
            logging.error(msg)
            exitUnlessIgnore(msg)
        elif oe.errno == 401:  # Unauthorized
            msg = "Authorization failure"
            logging.error(msg)
            exitUnlessIgnore(msg)
        elif oe.errno == 403:  # Forbidden
            msg = "Not Allowed"
            logging.error(msg)
            exitUnlessIgnore(msg)
        elif oe.errno == 409 and domain.endswith('/'):  # Conflict
            msg = "folder has sub-items"
            logging.error(msg)
            exitUnlessIgnore(msg)
        else:
            msg = f"Unexpected error: {oe}"
            logging.error(msg)
            exitUnlessIgnore(msg)
    if cfg["verbose"]:
        if domain.endswith('/'):
            msg = f"Folder: {domain} deleted"
        else:
            msg = f"Domain: {domain} deleted"
        cfg.print(msg)
           

#
# Usage
#
def usage():
    option_names = cfg.get_names()
    cmd = cfg.get_cmd()
    print("Usage:\n")
    print(f"    {cmd} [ OPTIONS ]  target")
    print("")
    print("Description:")
    print("    Delete given domains")
    print("       target: one or more domains to be deleted")
    print("")
    print("Options:")
    for name in option_names:
        help_msg = cfg.get_help_message(name)
        if help_msg:
            print(f"    {help_msg}")  
    print("")
    print("Examples:")
    print(f"     {cmd} /home/myfolder/file1.h5  /home/myfolder/file2.h5")
    print(f"     {cmd} hdf5://home/myfolder/file2.h5  hdf5://home/afolder/")
    print("")
    print(cfg.get_see_also(cmd))
    print("")
    sys.exit()

#
# Main
#
def main():

    cfg.setitem("help", False, flags=["-h", "--help"], help="this message")

    try:
        domains = cfg.set_cmd_flags(sys.argv[1:])
    except ValueError as ve:
        print(ve)
        usage()

    if len(domains) == 0:
        # need a domain
        usage()

    # setup logging
    logfname = cfg["logfile"]
    loglevel = cfg.get_loglevel()
    logging.basicConfig(filename=logfname, format='%(levelname)s %(asctime)s %(message)s', level=loglevel)
    logging.debug(f"set log_level to {loglevel}")

    for domain in domains:
        deleteDomain(domain)


if __name__ == "__main__":
    main()



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HSDS (HDF5 Scalable Data Service), Libraries and      #
# Utilities.  The full HSDS copyright notice, including                      #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

import sys
import logging

try:
    import h5py
    import h5pyd
except ImportError as e:
    sys.stderr.write("ERROR : %s : install it to use this utility...\n" % str(e))
    sys.exit(1)

if __name__ == "__main__":
    from config import Config
    from utillib import load_file
else:
    from .config import Config
    from .utillib import load_file

cfg = Config()  #  config object

#----------------------------------------------------------------------------------
def usage():
    option_names = cfg.get_names()
    cmd = cfg.get_cmd()
    print("Usage:\n")
    print(f"    {cmd} [ OPTIONS ]  domain [ filepath ]")
    print("")
    print("Description:")
    print("    Copy server domain to local HDF5 file")
    print("       domain: domain to be copied")
    print("       filepath: HDF5 file to be created ")
    print("")
    print("Options:")
    for name in option_names:
        help_msg = cfg.get_help_message(name)
        if help_msg:
            print(f"    {help_msg}")  
    print("")
    print("Examples:")
    print(f"     {cmd} /shared/tall.h5 tall.h5")
    print(f"     {cmd} hdf5://shared/tall.h5 tall.h5")
    print(f"     {cmd} hdf5://shared/tall.h5  # creates local file 'tall.h5'")
    print("")
    print(cfg.get_see_also(cmd))
    print("")
    sys.exit()

#end print_usage


#----------------------------------------------------------------------------------
def main():
    
    cfg.setitem("no_clobber", False, flags=["-n", "--no-clobber"],  help="do not overwrite target")
    cfg.setitem("nodata", False, flags=["--nodata",], help="do not copy dataset data")
    cfg.setitem("help", False, flags=["-h", "--help"], help="this message")

    try:
        cmdline_values = cfg.set_cmd_flags(sys.argv[1:])
    except ValueError as ve:
        print(ve)
        usage()

    if len(cmdline_values) < 1:
        usage()

    src_domain = cmdline_values[0]
    if len(cmdline_values) > 1:
        des_file = cmdline_values[1]
    else:
        # use domain base name as file
        parts = src_domain.split('/')
        des_file = parts[-1]

    # setup logging
    logfname = cfg["logfile"]
    loglevel = cfg.get_loglevel()
    logging.basicConfig(filename=logfname, format='%(levelname)s %(asctime)s %(message)s', level=loglevel)
    logging.debug(f"set log_level to {loglevel}")     

    logging.info(f"source domain: {src_domain}")
    logging.info(f"target file: {des_file}")

    # get a handle to input domain
    kwargs = {}
    kwargs["endpoint"] = cfg["hs_endpoint"]
    kwargs["username"] = cfg["hs_username"]
    kwargs["password"] = cfg["hs_password"]
    kwargs["bucket"] = cfg["hs_bucket"]
    try:
        fin = h5pyd.File(src_domain, mode='r', **kwargs)
    except IOError as ioe:
        if ioe.errno == 403:
            logging.error(f"No read access to domain: {src_domain}")
        elif ioe.errno == 404:
            logging.error(f"Domain: {src_domain} not found")
        elif ioe.errno == 410:
            logging.error(f"Domain: {src_domain} has been recently deleted")
        else:
            logging.error(f"Error opening domain {src_domain}: {ioe}")
        sys.exit(1)

    # create the output HDF5 file
    mode = "x" if cfg["no_clobber"] else "w"
    try:
        fout = h5py.File(des_file, mode)
    except IOError as ioe:
        logging.error(f"Error creating file {des_file}: {ioe}")
        sys.exit(1)

    try:
        kwargs = {}
        kwargs["verbose"] = cfg["verbose"]
        kwargs["ignore_error"] = cfg["ignore_error"]
        kwargs["dataload"] = None if cfg["nodata"] else "ingest"
        load_file(fin, fout, **kwargs)
        cfg.print(f"Domain {src_domain} downloaded to file: {des_file}")
    except KeyboardInterrupt:
        logging.error('Aborted by user via keyboard interrupt.')
        sys.exit(1)
#__main__
if __name__ == "__main__":
    main()



from setuptools import setup

# run:
#   setup.py install
# or (if you'll be modifying the package):
#   setup.py develop
# To use a consistent encoding
# To upload to PyPI:
# python setup.py sdist
# twine upload --repository-url https://upload.pypi.org/legacy/ dist/*
# twine upload dist/h5pyd-x.y.z.tar.gz
#
# Tag the release in github!
#
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="h5pyd",
    version="0.18.0",
    description="h5py compatible client lib for HDF REST API",
    long_description=long_description,
    url="http://github.com/HDFGroup/h5pyd",
    author="John Readey",
    author_email="jreadey@hdfgrouup.org",
    license="BSD",
    packages=["h5pyd", "h5pyd._hl", "h5pyd._apps"],
    install_requires=[
        "numpy >= 1.17.3",
        "requests_unixsocket",
        "pytz",
        "pyjwt",
        # "cryptography",
    ],
    setup_requires=["pkgconfig"],
    extras_require={
        "azure": [
            "msrestazure",
            "adal"
            ], 
        "google": [
            "google-api-python-client",
            "google-auth-oauthlib",
            "google-auth<2.0dev",],
        "aws": ["s3fs",],
        "hdf5": ["h5py",]
    },
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "hsinfo = h5pyd._apps.hsinfo:main",
            "hsls = h5pyd._apps.hsls:main",
            "hstouch = h5pyd._apps.hstouch:main",
            "hsacl = h5pyd._apps.hsacl:main",
            "hsdel = h5pyd._apps.hsdel:main",
            "hsrm = h5pyd._apps.hsdel:main",
            "hsget = h5pyd._apps.hsget:main",
            "hsload = h5pyd._apps.hsload:main",
            "hsconfigure = h5pyd._apps.hsconfigure:main",
            "hscopy = h5pyd._apps.hscopy:main",
            "hscp = h5pyd._apps.hscopy:main",
            "hsmv = h5pyd._apps.hsmv:main",
            "hsdiff = h5pyd._apps.hsdiff:main",
            "hsstat = h5pyd._apps.hsstat:main",
        ]
    },
)



import os
import sys
import json
import requests
import time
from abc import ABC, abstractmethod
from datetime import datetime

def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

# Azure
try:
    import adal
except ModuleNotFoundError:
    pass  # change this to the eprint below to see the import error
    # eprint()"Unable to import azure auth packages")

# Google
try:
    from google_auth_oauthlib.flow import InstalledAppFlow as GoogleInstalledAppFlow
    from google.auth.transport.requests import Request as GoogleRequest
    from google.oauth2.credentials import Credentials as GoogleCredentials
    from google.oauth2 import id_token as GoogleIDToken
except ModuleNotFoundError:
    pass  # change this to the eprint below to see the import error
    # eprint("Unable to import google auth packages")

from .config import Config

class OpenIDHandler(ABC):

    def __init__(self, endpoint, use_token_cache=True, username=None, password=None):
        """Initialize the token."""

        # Location of the token cache.
        self._token_cache_file = os.path.expanduser('~/.hstokencfg')
        self._endpoint = endpoint
        self._username = username
        self._password = password

        # The _token attribute should be a dict with at least the following keys:
        #
        # accessToken - The OpenID token to send.
        # refreshToken - The refresh token (optional).
        # expiresOn - The unix timestamp when the token expires (optional).

        if not use_token_cache or not os.path.isfile(self._token_cache_file):
            self._token = None
        else:
            if username:
                file_key = username + '@' + endpoint
            else:
                file_key = endpoint
            with open(self._token_cache_file, 'r') as token_file:
                self._token = json.load(token_file).get(file_key, None)

    @abstractmethod
    def acquire(self):
        """Acquire a new token from the provider."""
        pass

    @abstractmethod
    def refresh(self):
        """Refresh an existing token with the provider."""
        pass

    @property
    def username(self):
        """ Return username if known """
        return self._username

    @property
    def expired(self):
        """Return if the token is expired."""
        t = self._token
        # add some buffer to account for clock skew
        return t is not None and 'expiresOn' in t and time.time() + 10.0 >= t['expiresOn']

    @property
    def token(self):
        """Return the token if valid, otherwise get a new one."""

        if self.expired:
            self.refresh()
            if self._token:
                self.write_token_cache()

        if self._token is None:
            self.acquire()
            self.write_token_cache()

        return self._token['accessToken']

    def write_token_cache(self):
        """Write the token to a file cache."""

        cache_exists = os.path.isfile(self._token_cache_file)

        if self._username:
            file_key = self._username + '@' + self._endpoint
        else:
            file_key = self._endpoint

        # Create a new cache file.
        if not cache_exists and self._token is not None:
            with open(self._token_cache_file, 'w') as token_file:
                json.dump({file_key: self._token}, token_file)

        # Update an exisiting cache file.
        elif cache_exists:
            with open(self._token_cache_file, 'r+') as token_file:
                cache = json.loads(token_file.read())

                # Store valid tokens.
                if self._token is not None:
                    cache[file_key] = self._token

                # Delete invalid tokens.
                elif file_key in cache:
                    del cache[file_key]

                token_file.seek(0)
                token_file.truncate(0)
                json.dump(cache, token_file)


class AzureOpenID(OpenIDHandler):

    AUTHORITY_URI = 'https://login.microsoftonline.com'  # login endpoint for AD auth

    def __init__(self, endpoint, config=None):
        """Store configuration."""

        # Configuration manager
        hs_config = Config()

        # Config is a dictionary.
        if isinstance(config, dict):
            self.config = config

        # Maybe client_secrets are in environment variables?
        else:

            self.config = {
                'AD_APP_ID': hs_config.get("hs_ad_app_id", None),
                'AD_TENANT_ID': hs_config.get("hs_ad_tenant_id", None),
                'AD_RESOURCE_ID': hs_config.get("hs_ad_resource_id", None),
                'AD_CLIENT_SECRET': hs_config.get("hs_ad_client_secret", None)
            }

        if 'AD_CLIENT_SECRET' in self.config and self.config['AD_CLIENT_SECRET']:
            use_token_cache = False
        else:
            use_token_cache = True

        super().__init__(endpoint, use_token_cache=use_token_cache)

    def write_token_cache(self):
        if 'AD_CLIENT_SECRET' in self.config and self.config['AD_CLIENT_SECRET']:
            pass # don't use token cache for unattended authentication
        else:
            super().write_token_cache()

    def acquire(self):
        """Acquire a new Azure token."""

        if "adal" not in sys.modules:
            msg = "adal module not found, run: python setup.py install -e '.[azure]'"
            raise ModuleNotFoundError(msg)

        app_id = self.config["AD_APP_ID"]
        resource_id = self.config["AD_RESOURCE_ID"]
        tenant_id = self.config["AD_TENANT_ID"]
        client_secret = self.config.get("AD_CLIENT_SECRET", None)
        authority_uri = self.AUTHORITY_URI + '/' + tenant_id

        # Try to get a token using different oauth flows.
        context = adal.AuthenticationContext(authority_uri, enable_pii=True, api_version=None)

        try:
            if client_secret is not None:
                code = context.acquire_token_with_client_credentials(resource_id, app_id, client_secret)
            else:
                code = context.acquire_user_code(resource_id, app_id)

        except Exception as e:
            eprint("unable to process AD token: {}".format(e))
            self._token = None
            self.write_token_cache()
            raise

        if "message" in code:
            eprint(code["message"])
            mgmt_token = context.acquire_token_with_device_code(resource_id, code, app_id)

        elif "accessToken" in code:
            mgmt_token = code

        else:
            eprint("Could not authenticate with AD")

        # Only store some fields.
        self._token = {
            'accessToken': mgmt_token['accessToken'],
            'refreshToken': mgmt_token.get('refreshToken', None),
            'tenantId': mgmt_token.get('tenantId', tenant_id),
            'clientId': mgmt_token.get('_clientId', app_id),
            'resource': mgmt_token.get('resource', resource_id)
        }

        # Parse time to timestamp.
        if 'expiresOn' in mgmt_token:
            expire_dt = datetime.strptime(mgmt_token['expiresOn'], '%Y-%m-%d %H:%M:%S.%f')
            self._token['expiresOn'] = expire_dt.timestamp()

    def refresh(self):
        """Try to renew an Azure token."""

        try:

            # This will work for device code flow, but not with client
            # credentials. If we have the secret, we can just request a new
            # token anyways.

            authority_uri = self.AUTHORITY_URI + '/' + self._token['tenantId']
            context = adal.AuthenticationContext(authority_uri, api_version=None)
            mgmt_token = context.acquire_token_with_refresh_token(self._token['refreshToken'],
                                                                  self._token['clientId'],
                                                                  self._token['resource'],
                                                                  None)

            # New token does not have all the metadata.
            self._token['accessToken'] = mgmt_token['accessToken']
            self._token['refreshToken'] = mgmt_token['refreshToken']

            # Parse time to timestamp.
            if 'expiresOn' in mgmt_token:
                expire_dt = datetime.strptime(mgmt_token['expiresOn'], '%Y-%m-%d %H:%M:%S.%f')
                self._token['expiresOn'] = expire_dt.timestamp()

        except:
            self._token = None


class GoogleOpenID(OpenIDHandler):

    def __init__(self, endpoint, config=None, scopes=None):
        """Store configuration."""

        if "google.oauth2" not in sys.modules:
            msg = "google.oauth2 module not found, run: python setup.py install -e '.[google]'"
            raise ModuleNotFoundError(msg)
        
        # Configuration manager
        hs_config = Config()

        if scopes is None:
            scopes = hs_config.get('hs_google_scopes', 'openid').split()
        self.scopes = scopes

        # Config is a client_secrets dictionary.
        if isinstance(config, dict):
            self.config = config

        # Config points to a client_secrets.json file.
        elif isinstance(config, str) and os.path.isfile(config):
            with open(config, 'r') as f:
                self.config = json.loads(f.read())

        # Maybe client_secrets are in environment variables?
        else:
            self.config = {
                'installed': {
                    'project_id': hs_config.get('hs_google_project_id', None),
                    'client_id': hs_config.get('hs_google_client_id', None),
                    'client_secret': hs_config.get('hs_google_client_secret', None),
                    'auth_uri': 'https://accounts.google.com/o/oauth2/auth',
                    'token_uri': 'https://oauth2.googleapis.com/token',
                    'auth_provider_x509_cert_url': 'https://www.googleapis.com/oauth2/v1/certs',
                    'redirect_uris': ['urn:ietf:wg:oauth:2.0:oob', 'http://localhost']
                }
            }

        super().__init__(endpoint)

    def _parse(self, creds):
        """Parse credentials."""

        # NOTE: In Google OpenID, if a client is set up for InstalledAppFlow
        # then the client_secret is not actually treated as a secret. Acquire
        # will ALWAYS prompt for user input before granting a token.

        token = {
            'accessToken': creds.id_token,
            'refreshToken': creds.refresh_token,
            'tokenUri': creds.token_uri,
            'clientId': creds.client_id,
            'clientSecret': creds.client_secret,
            'scopes': creds.scopes
        }

        # The expiry field that is in creds is for the OAuth token, not the
        # OpenID token. We need to validate the OpenID tokenn to get the exp.
        idinfo = GoogleIDToken.verify_oauth2_token(creds.id_token, GoogleRequest())
        if 'exp' in idinfo:
            token['expiresOn'] = idinfo['exp']

        return token

    def acquire(self):
        """Acquire a new Google token."""

        flow = GoogleInstalledAppFlow.from_client_config(self.config,
                                                         scopes=self.scopes)
        creds = flow.run_console()
        self._token = self._parse(creds)

    def refresh(self):
        """Try to renew a token."""

        try:

            token = self._token
            creds = GoogleCredentials(token=None,
                                      refresh_token=token['refreshToken'],
                                      scopes=token['scopes'],
                                      token_uri=token['tokenUri'],
                                      client_id=token['clientId'],
                                      client_secret=token['clientSecret'])

            creds.refresh(GoogleRequest())
            self._token = self._parse(creds)

        except:
            self._token = None


class KeycloakOpenID(OpenIDHandler):

    def __init__(self, endpoint, config=None, scopes=None, username=None, password=None):
        """Store configuration."""

        # Configuration manager
        hs_config = Config()

        if scopes is None:
            scopes = hs_config.get('hs_keycloak_scopes', 'openid').split()
        self.scopes = scopes

        # Config is a client_secrets dictionary.
        if isinstance(config, dict):
            self.config = config

        # Config points to a client_secrets.json file.
        elif isinstance(config, str) and os.path.isfile(config):
            with open(config, 'r') as f:
                self.config = json.loads(f.read())

        # Maybe configs are in environment variables?
        else:
            self.config = {
                'keycloak_client_id': hs_config.get('hs_keycloak_client_id', None),
                'keycloak_client_secret': hs_config.get('hs_keycloak_client_secret', None),
                'keycloak_realm': hs_config.get('hs_keycloak_realm', None),
                'keycloak_uri': hs_config.get('hs_keycloak_uri', None)
            }

        super().__init__(endpoint, username=username, password=password)

    def _getKeycloakUrl(self):
        if not self.config['keycloak_uri']:
            raise KeyError("keycloak_uri not set")
        if not self.config['keycloak_realm']:
            raise KeyError("Keycloak realm not set")
        if not self.config['keycloak_client_id']:
            raise KeyError("keycloak client_id not set")

        url = self.config['keycloak_uri']
        url += "/auth/realms/" 
        url += self.config['keycloak_realm']
        url += "/protocol/openid-connect/token"

        return url

    def _parse(self, creds):
        """Parse credentials."""

        # validate json returned by keycloak
        if "token_type" not in creds:
            raise IOError("Unexpected Keycloak JWT, no token_type")
        if creds["token_type"].lower() != "bearer":
            raise IOError("Unexpected Keycloak JWT, expected Bearer token")

        token = {}
        if "access_token" not in creds:
            raise IOError("Unexpected Keycloak JWT, no access_token")
        token["accessToken"] = creds["access_token"]
        if "refesh_token" in creds:
            token["refreshToken"] = creds["refresh_token"]
        if "expires_in" in creds:
            now = time.time()
            token['expiresOn'] = now + creds["expires_in"]
       
        # TBD: client_secret
        # TBD: scopes
        # TBD: client_id

        return token

    def acquire(self):
        """Acquire a new Keycloak token."""
        keycloak_url = self._getKeycloakUrl()
        
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        body = {}
        body["username"] = self._username
        body["password"] = self._password
        body["grant_type"] = "password"
        body["client_id"] = self.config.get("keycloak_client_id")
        rsp = requests.post(keycloak_url, data=body, headers=headers)
        
        if rsp.status_code not in (200, 201):
            print("POST error: {}".format(rsp.status_code))
            raise IOError("Keycloak response: {}".format(rsp.status_code))

        creds = rsp.json()  # TBD: catch json format errors?
        self._token = self._parse(creds)

    def refresh(self):
        """Try to renew a token."""
        # TBD 
        # unclear if refresh is supported without a client secret
        self._token = None

